from enum import Enum
from dataclasses import dataclass
from typing import List

WHITESPACE = " \n\t"


class TokenType(Enum):
    KLEENE_CLOSURE = 0
    POSITIVE_CLOSURE = 1
    LEFT_PARENTHESIS = 2
    RIGHT_PARENTHESIS = 3
    CONCATENATION = 4
    UNION = 5
    SYMBOL = 6


@dataclass
class Token:
    type: TokenType
    value: any = None

    def __repr__(self):
        return self.type.name + (f":{self.value}" if self.value != None else "")


class Lexer:
    def __init__(self, expression) -> None:
        self.expression = iter(expression)
        self.title = self.forward()

    def forward(self):
        try:
            self.current_char = next(self.expression)
        except StopIteration:
            self.current_char = None

    def generate_tokens(self) -> List[Token]:
        while self.current_char != None:
            if self.current_char in WHITESPACE:
                self.forward()
                yield Token(TokenType.CONCATENATION)
            elif self.current_char == "⁺":
                self.forward()
                yield Token(TokenType.POSITIVE_CLOSURE)
            elif self.current_char == "*":
                self.forward()
                yield Token(TokenType.KLEENE_CLOSURE)
            elif self.current_char == "|":
                self.forward()
                yield Token(TokenType.UNION)
            elif self.current_char == "(":
                self.forward()
                yield Token(TokenType.LEFT_PARENTHESIS)
            elif self.current_char == ")":
                self.forward()
                yield Token(TokenType.RIGHT_PARENTHESIS)
            else:
                yield Token(TokenType.SYMBOL, self.current_char)
                self.forward()



from typing import List
from autome.regex import Token, TokenType

from autome.regex.nodes import (
    ConcatNode,
    KleeneClosureNode,
    ParserNode,
    PositiveClosureNode,
    SymbolNode,
    UnionNode,
)


class Parser:
    def __init__(self, tokens: List[Token]):
        self.tokens = iter(tokens)
        self.forward()

    def throw(self) -> None:
        problem = self.current_token
        rest = []

        while self.current_token is not None:
            self.forward()
            rest.append(self.current_token)

        raise Exception(f"Unexpected token {problem} in {rest}")

    def forward(self) -> None:
        try:
            self.current_token = next(self.tokens)
        except StopIteration:
            self.current_token = None

    def parse(self) -> ParserNode:
        if self.current_token == None:
            return None

        result = self.expression()

        if self.current_token != None:
            self.throw()

        return result

    def expression(self) -> ParserNode:
        result = self.term()

        while self.current_token != None and self.current_token.type in [
            TokenType.UNION
        ]:
            if self.current_token.type == TokenType.UNION:
                self.forward()
                result = UnionNode(result, self.term())

        return result

    def term(self) -> ParserNode:
        result = self.symbol()

        while self.current_token != None and self.current_token.type in [
            TokenType.CONCATENATION
        ]:
            if self.current_token.type == TokenType.CONCATENATION:
                self.forward()
                result = ConcatNode(result, self.term())

        return result

    def symbol(self) -> ParserNode:
        token = self.current_token

        # Opens new expression with parenthesis
        if token.type == TokenType.LEFT_PARENTHESIS:
            self.forward()

            expr = self.expression()

            if (
                self.current_token == None
                or self.current_token.type != TokenType.RIGHT_PARENTHESIS
            ):
                raise Exception("Expected ')'")

            self.forward()
            return self.unary_lookahead(expr)

        if token.type == TokenType.SYMBOL:
            self.forward()
            return self.unary_lookahead(SymbolNode(token.value))

        raise Exception("Unexpected token")

    def unary_lookahead(self, context) -> ParserNode:
        if (
            self.current_token is not None
            and self.current_token.type == TokenType.KLEENE_CLOSURE
        ):
            self.forward()
            return KleeneClosureNode(context)
        elif (
            self.current_token is not None
            and self.current_token.type == TokenType.POSITIVE_CLOSURE
        ):
            self.forward()
            return PositiveClosureNode(context)
        else:
            return context



from autome.automatas.finite_automata.machine import (
    DeterministicFiniteAutomata,
    NonDeterministicFiniteAutomata,
)
from autome.automatas.finite_automata.state import State
from autome.automatas.finite_automata.transition import Transition
from autome.automatas.finite_automata.parsers import JFlapConverter, JSONConverter



from autome.automatas.turing_machine.state import State
from autome.automatas.turing_machine.tape import Tape
from autome.automatas.turing_machine.transition import Transition
from typing import List, Callable, Deque
from pathlib import Path
from collections import deque
import json


class Machine:
    """
    Multitape deterministic Turing-Machine, accepts states, tapes and transitions as inputs.

    Outputs boolean for result upon transition to acceptance state or reject state.

    A machine can only have one initial state, but it may have multiple acceptance and reject states. Computation will be stopped as soon as the machine reaches one of those two types of state.
    """

    def __init__(
        self, states=[], tapes=[], transitions=[], branches: Deque = None
    ) -> None:
        self.tapes: List[Tape] = tapes
        self.states: List[State] = states
        self.current_state = next(filter(lambda state: state.initial, self.states))
        self.transitions: List[Transition] = transitions
        self.output = False

        if branches is not None:
            self.branches = branches
        else:
            self.branches: Deque[Machine] = deque([self])

    def __repr__(self):
        return f"Machine(states: {len(self.states)}, tapes: {len(self.tapes)})"

    @classmethod
    def parse(cls, path: Path) -> "Machine":
        """
        Creates an instance of Machine parsed from a json file located at @path that follows the schematics provided by the project (see /machines folder)

        Throws FileNotFoundError if there's no file at @path
        Throws ValueError if the json file doesn't match the schematics
        """
        with open(path, "r", encoding="utf8") as file:
            model = json.loads(file.read())
            tapes = [Tape.parse(tape) for tape in model["tapes"]]
            states = [State.parse(state) for state in model["states"]]
            transitions = [
                Transition.parse(transition, states)
                for transition in model["transitions"]
            ]

        return Machine(states=states, tapes=tapes, transitions=transitions)

    def clone(self) -> "Machine":
        """
        Returns a clone of the calling Turing Machine. The states, transitions and branches will be shallow copies
        while the tapes are clones with differente memory spaces.

        States and Transitions are read-only, so a shallow copy is fine.

        Tapes have write and move operations, so they must be new objects on the new machine.

        The branches must be shared between instances (we call it a "gambiarra")
        """
        tapes = [tape.clone() for tape in self.tapes]

        return Machine(
            states=self.states,
            tapes=tapes,
            transitions=self.transitions,
            branches=self.branches,
        )

    def execute_transition(self, transition: Transition) -> "Machine":
        """
        Executes a transition into the machine, writing and moving the tapes to apply changes.
        """
        for i in range(len(self.tapes)):
            self.tapes[i].write(transition.writes[i])
            self.tapes[i].move(transition.moves[i])

        self.current_state = transition.destiny

        return self

    def accepts(self, inputs: List[str], steps=False) -> bool:
        """
        Resets the Turing Machine to an entry state and runs the computation for a list of inputs that will be inserted into
        the machine's tapes.
        """
        self.tapes = [Tape(value) for value in inputs]
        self.current_state = next(filter(lambda state: state.initial, self.states))
        self.branches: Deque[Machine] = deque([self])

        return self.run(steps)

    def run(self, steps=False) -> bool:
        """
        Runs the computation for the current Turing-Machine.
        If @steps is True, computation will run in a step-by-step mode, use this for debug. To control step by step mode, hit Enter on REPL to go to the next step.
        """
        output = False
        while len(self.branches) > 0:
            branch = self.branches.popleft()
            if steps:
                input()
            output = branch.step()
        return output

    def step(self) -> bool:
        """
        Runs one step of the computation for the current Turing-Machine.

        The step will take the current input of all the machine's tapes, compare with the list
        of available transitions trying to find a match. If it doesn't find, the machine will
        be pushed into a rejection state, where the computation will stop.

        However, if it does find a match (and only one, because the machine stils can't handle non-determinism)
        it will apply all the changes requested on tapes, then shifts the current state of the machine
        to the destiny of the found transition.
        """
        symbols = [tape.get() for tape in self.tapes]

        condition: Callable[[Transition], bool] = (
            lambda transition: symbols == transition.reads
            and self.current_state == transition.origin
        )

        matches = list(filter(condition, self.transitions))

        if len(matches) == 0:
            return self.current_state.accept
        else:
            for transition in matches[1:]:
                self.branches.append(self.clone().execute_transition(transition))

            self.branches.append(self.execute_transition(matches[0]))



from autome.automatas.finite_automata.machine import NonDeterministicFiniteAutomata
from autome.automatas.finite_automata.state import State
from autome.automatas.finite_automata.transition import Transition


class SymbolAutomata(NonDeterministicFiniteAutomata):
    """Creates a symbol recognizer automata, which accepts the symbol @symbol.

        The transition table of the automata is:
        | states |   by   |
        | → q0   | symbol |
        | * q1   |   --   |

    Args:
        symbol (string): the symbol to be recognized.
    """

    def __init__(self, symbol) -> "SymbolAutomata":
        states = [State(initial=True), State(accept=True)]
        transitions = [Transition(states[0], states[1], symbol)]
        title = f"Reconhecedor da ER: {symbol}"
        super().__init__(states, transitions, title)



import json
from autome.utils.enums import Direction


class Tape:
    """
    Computational representation of an infinite tape of memory
    Cause there's no such thing as infinite memory we're using a statically sized list with a pointer
    Tapes may only contain a single char for position, so any input passed as @initial_value will be splitted that way.
    """

    def __init__(self, initial_value: str = "", pointer: int = 0):
        self.pointer = pointer
        self.data = list(initial_value)

    def __repr__(self):
        return "  " + " " * (self.pointer * 5) + "↓\n" + json.dumps(self.data)

    def __len__(self):
        return len(self.data)

    @classmethod
    def parse(cls, model: dict) -> "Tape":
        """
        Returns a new instance of Tape based on the contents of @model. This function was written to be used within Machine.parse
        """
        return Tape(initial_value=model["initial_string"])

    def get(self):
        """
        Returns the content of the tape at the pointer's position
        """
        if self.pointer < 0 or self.pointer >= len(self):
            return "_"
        else:
            return self.data[self.pointer]

    def write(self, value: str):
        """
        Writes the content passed in @value to the tape at the pointer's position
        """
        if self.pointer < len(self):
            self.data[self.pointer] = value
        else:
            self.data.append(value)

    def move(self, direction: Direction):
        """
        Moves the pointer to the left or the right
        """
        if direction == Direction.LEFT:
            self.pointer -= 1
        elif direction == Direction.RIGHT:
            self.pointer += 1

    def clone(self) -> "Tape":
        """
        Returns a clone of the calling tape
        """
        return Tape("".join(self.data), pointer=self.pointer)



import json
from pprint import pprint
import click

from typing import List
from tabulate import tabulate
from autome.utils.dataclasses import Token
from autome.utils.errors import SyntaxException


class CFG:
    def __init__(
        self,
        nonterminals: List[str],
        terminals=List[str],
        initial=str,
        productions=None,
        name="",
    ):
        self.nonterminals = nonterminals if nonterminals else list()
        self.terminals = terminals if terminals else list()
        self.initial = initial if initial else list()
        self.productions = productions if productions else dict()
        self.name = name

        self.first = dict()
        self.follow = dict()

    def __repr__(self) -> str:
        content = []
        for head, body in self.productions.items():
            prods = ["".join(prod) for prod in body]
            joined = "|".join(prods)
            content.append(f"{head} → {joined}\n")

        return "".join(content)

    @classmethod
    def parse(cls, input: click.Path):
        with open(input, "r") as file:
            content = json.load(file)

        grammar = content["grammar"]

        nonterminals = grammar["nonterminals"]

        initial = grammar["initial"]
        if initial not in nonterminals:
            raise Exception("The initial symbol must be a part of the nonterminals set")

        terminals = grammar["terminals"]

        productions = dict()

        for production in grammar["productions"]:
            prods = [prod.split(" ") for prod in production["body"].split("|") if prod]
            prods = [list(filter(lambda val: val != "", prod)) for prod in prods]
            productions[production["head"]] = prods

        return CFG(nonterminals, terminals, initial, productions)

    def calculate_first(self):
        """Calculate the 'first' set of all the symbols in the grammar

        Returns:
            Dict: mapping of the sets containing the 'first', indexed by symbol
        """
        self.first.clear()

        # First(a) = {a}
        for terminal in self.terminals:
            self.first[terminal] = {terminal}

        # Para cada não terminal X
        for nonterminal in self.nonterminals:
            self.first[nonterminal] = set()

            # Se X ::= aY, a pertence à First(X) = First(X) ∪ {a}
            for production in self.productions[nonterminal]:
                if production[0] in self.terminals:
                    self.first[nonterminal].add(production[0])

        # 2. Se X ::= Y1 Y2...Yk, então FIRST(Y1) pertence à FIRST(X)
        stop = False
        while not stop:
            stop = True
            for nonterminal in self.nonterminals:
                for production in self.productions[nonterminal]:
                    for symbol in production:
                        if symbol in self.nonterminals:
                            firstSymbol = self.first[symbol].copy()
                            if "&" in firstSymbol:
                                firstSymbol.remove("&")
                            if not firstSymbol.issubset(self.first[nonterminal]):
                                self.first[nonterminal] = self.first[nonterminal].union(
                                    firstSymbol
                                )
                                stop = False

                            if "&" not in self.first[symbol]:
                                break
                        else:
                            self.first[nonterminal].add(symbol)
                            break

        return self.first

    def first_of_sequence(self, sequence):
        """Return the first set of a sequence (calculating nullable symbols)

        Args:
            sequence List[str]: the input sequence

        Returns:
            Set[str]: the set holding the firsts of the sequence
        """
        firsts = set()

        for index, symbol in enumerate(sequence):
            firsts = firsts.union(self.first[symbol]) - {"&"}

            # Se o first não tem  epsilon, não precisamos continuar pensando se será ou não anulável
            if "&" not in self.first[symbol]:
                break

            # Se chegou no último símbolo da produção, e esse símbolo é anulável, então precisamos adicionar o follow também
            if index == len(sequence) - 1:
                firsts.add("&")

        return firsts

    def calculate_follow(self):
        """Calculate the follow set for all the nonterminals in the grammar

        Returns:
            Dict: mapping with all the follow sets, indexed by nonterminals
        """
        self.follow.clear()

        self.calculate_first()

        follows = {symbol: set() for symbol in self.nonterminals}

        follows[self.initial].add("$")

        changed = True

        while changed:
            changed = False

            for head, productions in self.productions.items():
                for production in productions:
                    for i, symbol in enumerate(production):
                        if symbol in self.terminals:
                            continue

                        sequence = production[i + 1 :]

                        # This means that this nonterminal is the last in the production
                        if len(sequence) == 0:
                            # if production == ['&']:
                            #     continue

                            follows[symbol] = follows[symbol].union(follows[head])

                            continue

                        firsts = self.first_of_sequence(sequence)

                        old_follow = follows[symbol].copy()

                        follows[symbol] = follows[symbol].union(firsts - {"&"})

                        if "&" in firsts:
                            follows[symbol] = follows[symbol].union(follows[head])

                        if old_follow != follows[symbol]:
                            changed = True

        self.follow = follows

        return self.follow

    def eliminate_direct_left_recursion(self, symbol):
        """Helper to eliminate direct left recursion on a given symbol

        Args:
            symbol String: the non terminal holding the left recursion
        """
        alphas = list()
        betas = list()
        newSymbol = f"{symbol}'"

        for production in self.productions[symbol]:
            if production[0] is symbol:
                alpha = production[1:].copy()
                alpha.append(newSymbol)
                alphas.append(alpha)
            elif production == ["&"]:
                betas.append(["&"])
            else:
                beta = production.copy()
                beta.append(newSymbol)
                betas.append(beta)

        if len(alphas) > 0:
            alphas.append(["&"])
            self.productions[newSymbol] = alphas
            self.nonterminals.append(newSymbol)
            self.productions[symbol] = betas

    def eliminate_left_recursion(self):
        """Eliminate left recursion by manipulating the grammar and passing the recursion to the right of the productions

        Returns:
            Grammar: the altered grammar
        """

        for symbol in self.nonterminals:
            self.eliminate_direct_left_recursion(symbol)

        nonterminals = enumerate(self.nonterminals)
        for (i, nonTerminali) in nonterminals:
            for nonTerminalj in [
                nonTerminalj for j, nonTermninalj in nonterminals if j < i
            ]:
                for production in self.productions[nonTerminali]:
                    if production[0] is nonTerminalj:
                        self.productions[nonTerminali].remove(production)
                        for productionBeta in self.productions[nonTerminalj]:
                            productionBeta = productionBeta.copy()
                            productionBeta.append(production[1:].copy())
                            self.productions[nonTerminali].append(productionBeta)
            self.eliminate_direct_left_recursion(nonTerminali)

        return self

    def left_factoring(self, *, iters=10):
        """Procedure for removing non-determinism on the grammar. While removing direct determinism, you can create
        a new non-determinism in another production and this can occur in a loop, so this proccess is not decidible
        by a Turing Machine. Hence, this function tries up to @iters times to remove, if after this number of
        iterations the grammar keeps changing, we return false and you should stop the execution of your program

        Args:
            iters (int, optional): Total iterations. Defaults to 10.

        Returns:
            boolean: true if the grammar could be factorated, false otherwise.
        """
        self.remove_direct_non_determinism()

        for _ in range(iters):
            changed = self.remove_indirect_non_determinism()

            self.remove_direct_non_determinism()

            if not changed:
                break
        else:
            return False

        return True

    def derive(self, prod):
        """Recursively returns a list of all derivations that can be made by a given production

        Args:
            prod List[string]: list of symbols representing a production

        Returns:
            List[List[string]]: all the derivations that can be generated starting at the given production
        """
        if not prod:
            return [[]]

        prod_ = prod[0]

        if prod_ in self.terminals:
            return [[prod_] + derivation for derivation in self.derive(prod[1:])]

        elif prod_ in self.productions:
            out = []
            derivations = self.derive(prod[1:])
            for p in self.productions[prod_]:
                if p == ["&"]:
                    out += derivations
                else:
                    if derivations:
                        out += [p + deriv for deriv in derivations]
                    else:
                        out += p
            return out

    def remove_direct_non_determinism(self):
        """Searches for direct non-determinism and removes them by factoring the grammar to the left"""
        productions = self.productions
        new = {}

        for nonterminal in productions:
            new[nonterminal] = []
            prefixes = []

            for first in productions[nonterminal]:
                if not prefixes:
                    prefixes.append(first)
                    continue

                prefix_found = False
                prefix = []

                for i, second in enumerate(prefixes):
                    for p1, p2 in zip(first, second):
                        if p1 != p2:
                            break
                        prefix.append(p1)
                    if prefix and not prefix_found:
                        prefixes[i] = prefix
                        prefix_found = True
                if not prefix_found:
                    prefixes.append(first)

            count = 1
            for pref in prefixes:
                aux = []
                for prod in productions[nonterminal]:
                    if len(pref) <= len(prod) and pref == prod[: len(pref)]:
                        aux.append(prod)
                if len(aux) > 1:
                    symbol = f"{nonterminal}" + ("'" * count)
                    count += 1
                    self.nonterminals.append(symbol)
                    new_prod = pref + [symbol]

                    if new_prod not in new[nonterminal]:
                        new[nonterminal].append(new_prod)

                    new[symbol] = []

                    for p in aux:
                        p = p[len(pref) :]
                        p = p if len(p) > 0 else ["&"]
                        new[symbol].append(p)
                else:
                    new[nonterminal].append(pref)

        self.productions = new

    def remove_indirect_non_determinism(self):
        """Search for indirect non-determinisms and removes them by transforming into direct
        nondeterminisms using sucessive derivations

        Returns:
            boolean: true if the function modified the grammar, false otherwise
        """

        def get_firsts_chain(chain):
            """Returns the first set of an entire sequence of symbols

            Args:
                chain List[String]: the chain for which the first should be calculated

            Returns:
                List[String]: A list containing First(@chain)
            """
            self.calculate_first()

            if chain[0] == "&":
                return ["&"]

            return list(self.first[chain[0]])

        changed = False

        for symbol in self.productions:
            aux = []
            worrisome = set()

            for prod in self.productions[symbol]:
                firsts = get_firsts_chain(prod)

                for a in aux:
                    if len(set(firsts).intersection(a[1])) > 0:
                        worrisome.add(tuple(prod))
                        worrisome.add(tuple(a[0]))
                        changed = True

                for i, symbol in enumerate(prod[:-1]):
                    if symbol in self.nonterminals and "&" in self.first[symbol]:
                        if (
                            len(
                                self.first[symbol].intersection(
                                    get_firsts_chain(prod[i + 1 :])
                                )
                            )
                            > 0
                        ):
                            try:
                                worrisome.add(prod)
                            except TypeError:
                                print(f"Opsie {prod}")
                            changed = True

                aux.append((prod, firsts))

            for prod in worrisome:
                self.productions[symbol].remove(list(prod))

            for prod in worrisome:
                derivations = self.derive(list(prod))
                for d in derivations:
                    if d not in self.productions[symbol]:
                        self.productions[symbol].append(d)
        return changed

    def table(self):
        """Mounts the syntax analysis table for the LL(1) algorithim using first and follow.

        Returns:
            Dict: the table containing all the information needed for LL(1) parsing
        """
        self.calculate_first()
        self.calculate_follow()

        table = {}

        for production_head, productions in self.productions.items():
            table[production_head] = {}

            for alpha in productions:
                firsts = set()

                # Edge case for a production like B -> &
                if alpha[0] == "&":
                    firsts = firsts.union(self.follow[production_head])

                # Verificar se a produção pode ser anulável
                for symbol in alpha:
                    firsts = firsts.union(self.first[symbol]) - {"&"}

                    # Se o first não tem  epsilon, não precisamos continuar pensando se será ou não anulável
                    if "&" not in self.first[symbol]:
                        break

                    # Se chegou no último símbolo da produção, e esse símbolo é anulável, então precisamos adicionar o follow também
                    if symbol == alpha[-1]:
                        firsts = firsts.union(self.follow[production_head])

                for symbol in firsts:
                    table[production_head][symbol] = alpha

        return table

    def display_analysis_table(self, table):
        print("\nLL(1) analysis table\n")
        m = len(self.terminals) + 2
        n = len(self.nonterminals) + 2

        data = [["-" for i in range(m)] for j in range(n)]
        for column in range(1, n):
            data[column][0] = self.nonterminals[column - 2]

        for row in range(1, m):
            if row == len(self.terminals) + 1:
                data[0][row] = "$"
            else:
                data[0][row] = self.terminals[row - 2]

        for column in range(1, n):
            for row in range(1, m):
                if data[0][row] == "$":
                    nonterminal = data[column][0]
                    terminal = "$"
                else:
                    nonterminal = data[column][0]
                    terminal = data[0][row]

                try:
                    data[column][row] = " ".join(table[nonterminal][terminal])
                except KeyError:
                    pass

        print(tabulate(data, tablefmt="fancy_grid"))

    def accept(self, tokens: List[Token], debug=False):
        """Validates a sequence of tokens using the LL(1) parser. This wrapper already
        calculate First and Follow sets, eliminates left recursion and apply left factoring
        to remove non-determinism to the grammar.

        Args:
            tokens (List[Token]): List of tokens to be validated
            debug (bool, optional): Debug level flag. Defaults to False.

        Raises:
            SyntaxException: if a syntax error is found

        Returns:
            boolean: the result of the validation
        """
        self.calculate_first()
        self.calculate_follow()
        self.eliminate_left_recursion()
        self.left_factoring()

        table = self.table()
        self.display_analysis_table(table)
        stack = ["$", self.initial]

        sentence = " ".join([token.type for token in tokens])
        sentence = sentence.split(" ")
        sentence.append("$")
        i = 0
        symbol = sentence[i]

        if debug:
            print("\nLL(1) recognition stack:\n")

        while stack != ["$"] and i < len(sentence):
            if debug:
                print(f"Input: {symbol}; Stack: {stack}")

            if stack[-1] in self.terminals:
                if i + 1 < len(sentence):
                    i += 1

                symbol = sentence[i]
                stack.pop()

                continue

            if symbol not in table[stack[-1]]:
                if symbol == "$":
                    raise SyntaxException(f"Unexpected EOL: {sentence[i - 1]}")
                else:
                    raise SyntaxException(f"Unexpected symbol: {symbol}")

            prod = table[stack[-1]][symbol]
            stack.pop()

            if prod == ["&"]:
                continue
            elif prod:
                for p in reversed(prod):
                    stack.append(p)

        if debug:
            print(f"Input: {symbol}; Stack: {stack}")

        return stack == ["$"] and sentence[i] == "$"



from copy import deepcopy
import pdb
from tabulate import tabulate
from typing import Callable, Dict, List, Set, Tuple
from autome.automatas.finite_automata.state import State
from autome.automatas.finite_automata.transition import Transition
from autome.base_machine import BaseMachine


class DeterministicFiniteAutomata(BaseMachine):
    """
    Finite automata, accepts states and transitions as inputs.

    An automata can only have one initial state, but it may have multiple acceptance states.
    """

    def __init__(
        self, states=[], transitions=[], title="FiniteAutomata", description=""
    ) -> None:
        self.states: List[State] = list(states)
        self.transitions: List[Transition] = list(transitions)
        self.title = title
        self.description = description
        self.step_stack = []
        self.create_transition_map()

    def create_transition_map(self):
        self.transition_map: Dict[State, Dict[str, Set[str]]] = dict()

        for state in self.states:
            if state not in self.transition_map:
                self.transition_map[state] = {}

            for transition in self.transitions:
                if transition.origin == state:
                    if transition.symbol not in self.transition_map[state]:
                        self.transition_map[state][transition.symbol] = set()
                    self.transition_map[state][transition.symbol].add(
                        transition.destiny
                    )

    def add_transition(self, origin: State, destiny: State, symbol: str) -> None:
        """Adds a new transition the the automata

        Args:
            origin (State): state from which the transition starts
            destiny (State): state where the transition ends
            symbol (str): trigger symbol of the transition
        """

        self.transitions = list(
            set(
                self.transitions
                + [Transition(origin=origin, destiny=destiny, symbol=symbol)]
            )
        )

    def add_state(self, state: State) -> None:
        """Adds a new state the the automata

        Args:
            state (State): new state
        """

        self.states = list(set(self.states + [state]))

    def execute_transition(self, transition: Transition) -> None:
        """
        Executes a transition into the automata, moving reading head to the right and applying changes to the current state.
        """
        self.step_stack.append(transition)
        self.current_state = transition.destiny

    def final(self) -> List[State]:
        return list(filter(lambda state: state.accept, self.states))

    def initial(self) -> State:
        return next(filter(lambda state: state.initial, self.states))

    def accepts(self, word: str, debug=False) -> bool:
        """
        Resets the finite automata to an initial state and runs the computation for a given word.
        """
        self.current_state = self.initial()

        if debug:
            print(f"Starting at {self.current_state}")

        for character in word:
            result = self.step(character)
            if not result:
                return False
            if debug:
                print(
                    f"Transition to {self.current_state} by {self.step_stack[-1].symbol}"
                )

        return self.current_state.accept

    def step(self, character) -> bool:
        """
        Runs one step of the computation for the current finite automata.

        The step will take the current symbol from the reading head, compare with the list
        of available transitions trying to find a match. If it doesn't find, the machine will
        be pushed into a rejection state, where the computation will stop.

        However, if it does find a match shifts the current state of the machine
        to the destiny of the found transition and moves the reading head.
        """

        condition: Callable[[Transition], bool] = (
            lambda transition: character == transition.symbol
            and self.current_state == transition.origin
        )

        matches = list(filter(condition, self.transitions))

        if len(list(matches)) == 0:
            return False
        else:
            self.execute_transition(matches[0])
            return True

    def complement(self) -> "DeterministicFiniteAutomata":
        """Generate the complement of a given automata. The algorithm is pretty straight-forward, just note
        that we had to use Python's deepcopy to create a new object exactly equal to the old one but with different
        memory address

        Returns:
            DeterministicFiniteAutomata: a new automata, representing the complement of the operand
        """
        new = deepcopy(self)

        for state in new.states:
            state.accept = not state.accept

        return new

    def union(
        self, other: "DeterministicFiniteAutomata"
    ) -> "DeterministicFiniteAutomata":
        """Generate an union between the AF, the original algorithm returns a NDFA, but for use purposes
        we're determinizing the resulting machine before returning

        Returns:
            DeterministicFiniteAutomata: a new automata, representing the union between operands
        """

        old_initials = (
            self.initial(),
            other.initial(),
        )  # Old initial states from self and other
        old_finals = (
            self.final() + other.final()
        )  # Old final states from self and other

        new_final = State("qnf", accept=True)
        new_initial = State("qni", initial=True)

        states = [new_final, new_initial] + self.states + other.states

        for state in states:
            if state != new_initial:
                state.initial = False

        # From the new initial state, creates an epsilon transition to all the old initial states
        new_to_old_initial = [
            Transition(new_initial, old_initial, "&") for old_initial in old_initials
        ]

        # From the old final states, creates and epsilon transition to the new final state
        old_to_new_final = [
            Transition(old_final, new_final, "&") for old_final in old_finals
        ]

        transitions = (
            self.transitions + other.transitions + new_to_old_initial + old_to_new_final
        )

        # Because we have epsilon transitions, the resultant is a NDFA
        # that should be determinized before used
        new = NonDeterministicFiniteAutomata(states=states, transitions=transitions)
        new.create_transition_map()

        return new.determinize()

    def cross_union(
        self, other: "DeterministicFiniteAutomata"
    ) -> "DeterministicFiniteAutomata":
        result = NonDeterministicFiniteAutomata()

        initial = State("q0", initial=True)
        result.add_state(initial)

        state_mapping = {}
        index = 1

        for state in sorted(self.states):
            state_mapping[state] = State(f"q{index}", accept=state.accept)
            result.add_state(state_mapping[state])
            index += 1

        result.create_transition_map()

        symbols = list(set(map(lambda transition: transition.symbol, self.transitions)))
        symbols.sort()

        for state in sorted(self.states):
            for symbol in symbols:
                if symbol in self.transition_map[state]:
                    destiny_states = self.transition_map[state][symbol]
                    for destiny_state in destiny_states:
                        result.add_transition(
                            state_mapping[state], state_mapping[destiny_state], symbol
                        )

        result.add_transition(initial, state_mapping[self.initial()], "&")

        state_mapping = {}

        for state in sorted(other.states):
            state_mapping[state] = State(f"q{index}", accept=state.accept)
            result.add_state(state_mapping[state])
            index += 1

        result.create_transition_map()

        symbols = list(set(map(lambda transition: transition.symbol, self.transitions)))
        symbols.sort()

        for state in sorted(other.states):
            for symbol in symbols:
                if symbol in other.transition_map[state]:
                    destiny_states = other.transition_map[state][symbol]
                    for destiny_state in destiny_states:
                        result.add_transition(
                            state_mapping[state], state_mapping[destiny_state], symbol
                        )

        result.add_transition(initial, state_mapping[other.initial()], "&")

        result.create_transition_map()

        # pdb.set_trace()

        return result.determinize()

    def intersection(
        self, other: "DeterministicFiniteAutomata"
    ) -> "DeterministicFiniteAutomata":
        """Generate an intersection of the given automatas, the algorithm actually uses other
        automata operation's to mount the new machine.

        Returns:
            DeterministicFiniteAutomata: a new automata, representing the intersection between operands
        """
        a = self.complement()
        b = other.complement()

        result = (a | b).complement()

        for state in result.states:
            if state.accept and state.initial:
                state.accept = False

        return result

    def clone(self) -> "DeterministicFiniteAutomata":
        mapping = {}

        new_states = []
        new_transitions = []

        for original in self.states:
            new = State(
                accept=original.accept, initial=original.initial, type=original.type
            )
            mapping[original] = new
            new_states.append(new)

        for original in self.transitions:
            try:
                new = Transition(
                    mapping[original.origin], mapping[original.destiny], original.symbol
                )
                new_transitions.append(new)
            except:
                pass

        return DeterministicFiniteAutomata(new_states, new_transitions)

    def minimize(self) -> "DeterministicFiniteAutomata":
        for state in self.states:
            if not state.accept and len(self.transition_map[state]) == 0:
                print(f"{state} is dead")

        visited = set()

        for transition in self.transitions:
            visited.add(transition.destiny)

        visited.add(self.initial())

        mapping = {}
        states = []
        transitions = []

        for state in visited:
            new = State(accept=state.accept, initial=state.initial, type=state.type)
            mapping[state] = new
            states.append(new)

        for transition in self.transitions:
            try:
                new = Transition(
                    mapping[transition.origin],
                    mapping[transition.destiny],
                    transition.symbol,
                )
                transitions.append(new)
            except:
                pass

        return DeterministicFiniteAutomata(states, transitions)

    # Dunder methods to allow operator overloading
    def __or__(self, other):
        return self.cross_union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __invert__(self):
        return self.complement()

    def __repr__(self):
        symbols = list(set(map(lambda transition: transition.symbol, self.transitions)))
        symbols.sort()

        headers = ["state"] + list(symbols)

        data = []

        self.states.sort()

        for state in self.states:
            name = state.name

            if state.initial:
                name = "→" + name

            if state.accept:
                name = name + "*"

            if state.type:
                name = name + f" ({state.type})"

            aux = [name]

            for symbol in symbols:
                if symbol not in self.transition_map[state]:
                    aux.append("-")
                else:
                    aux.append(self.transition_map[state][symbol])
            data.append(aux)

        return tabulate(data, headers=headers, tablefmt="fancy_grid")

    def __eq__(self, other: "DeterministicFiniteAutomata") -> bool:
        for state in self.states:
            if state not in other.states:
                return False

        for transition in self.transitions:
            if transition not in other.transitions:
                return False

        return True


class NonDeterministicFiniteAutomata(DeterministicFiniteAutomata):
    def run(self, word):
        return self.determinize().run(word)

    def e_closure(self, states: List[State]) -> Tuple[Set[State], bool]:
        """Calculates the sigma-closure of a given list of states in the current automata
        also returning if the resulting sigma-closure is a new final state or not

        Args:
            states (List[State]): the states that will be merged

        Returns:
            Set[State]: sigma-closure set, containing all the states reached by &-transitions
            from the original states passed
        """
        closure = set()
        accept = False

        while states:
            state = states.pop()
            closure.add(state)

            if state.accept:
                accept = True

            if "&" in self.transition_map[state]:
                for destiny in self.transition_map[state]["&"]:
                    if destiny not in closure:
                        states.append(destiny)

        return (closure, accept)

    def determinize(self) -> "DeterministicFiniteAutomata":
        """Determinizes a NDFA, returning an equivalent DFA

        Returns:
            DeterministicFiniteAutomata: the equivalente DFA to the operand
        """
        symbols = list(set([transition.symbol for transition in self.transitions]))

        new = DeterministicFiniteAutomata()

        # Cria uma pilha com todos os estados
        state_stack = [{state} for state in self.states]

        # Enquanto possuir algo na pilha
        while state_stack:
            # Desempilha uma lista?
            states = list(state_stack.pop())

            # Calcula o sigma fecho dessa lista de estados
            (state_closure, _) = self.e_closure(states)

            # Cria um novo estado, com o join de todos os estados no sigma fecho
            closure_state = State.join(list(state_closure))

            # Adiciona o novo estado na lista de estados do autômato determinístico
            new.add_state(closure_state)

            # Pra cada simbolo do alfabeto
            for symbol in symbols:
                # Pula epsilon
                if symbol == "&":
                    continue

                reached_from_closure = set()

                # Pra cad estado no fecho original
                for to_state in state_closure:
                    # Se não existe nenhuma transição por esse estado, pula
                    if symbol not in self.transition_map[to_state]:
                        continue
                    # Se existe, atualiza o conjunto anterior de transições com as transições
                    # do autômato atual que partam do conjunto atual dentro do fecho, pelo simbolo atual
                    reached_from_closure.update(self.transition_map[to_state][symbol])

                # Se o conjunto estiver vazio, pula o simbolo
                if len(reached_from_closure) == 0:
                    continue

                # Calcula um novo fecho
                (closure, _) = self.e_closure(list(reached_from_closure))

                # Cria um novo estado a partir da união do novo fecho
                new_state = State.join(list(closure))

                # Se o novo estado criado não estiver na pilha, nem nos estados atuais, adiciona na pilha
                if not (new_state in new.states) and not (new_state in state_stack):
                    state_stack.append(closure)

                # Adiciona uma transição do estado fecho para o novo estado, pelo símbolo
                new.add_transition(closure_state, new_state, symbol)

        new.create_transition_map()

        return new



from autome.automatas.finite_automata.machine import (
    NonDeterministicFiniteAutomata,
)
from autome.automatas.finite_automata.state import State
from autome.automatas.finite_automata.transition import Transition


class KleeneAutomata(NonDeterministicFiniteAutomata):
    """Creates a concatenation recognizer automata, which accepts the Kleene closure of the languages regognized by its inputs.

    Args:
        a (NonDeterministicFiniteAutomata): first operand.
    """

    def __init__(self, a: NonDeterministicFiniteAutomata) -> "KleeneAutomata":
        # Using deep copys to avoid modifying the original automata.
        temp_a = a.clone()

        # Creating new initial and final state
        new_initial = State(initial=True)
        new_final = State(accept=True)

        new_transitions = []

        # Creating epsilon transitions between the new initial and old initial states
        new_transitions.append(Transition(new_initial, new_final, "&"))
        new_transitions.append(Transition(new_initial, temp_a.initial(), "&"))

        # For each final state we should create epsilon transitions to the old initial state and to the new final state
        for state in temp_a.final():
            new_transitions.append(Transition(state, temp_a.initial(), "&"))
            new_transitions.append(Transition(state, new_final, "&"))
            state.accept = False

        # Marking old initial states as non-initial
        temp_a.initial().initial = False

        states = temp_a.states + [new_initial, new_final]
        transitions = temp_a.transitions + new_transitions

        super().__init__(states, transitions)



from autome.utils.enums import Direction
from os import read
from autome.automatas.turing_machine.tape import Tape
from autome.automatas.turing_machine.transition import Transition
from autome.automatas.turing_machine.state import State
from pathlib import Path
from xml.etree import ElementTree as ET
from autome.automatas.turing_machine.machine import Machine
from typing import Callable, List


class JFlapConverter:
    @classmethod
    def parse(cls, source: Path) -> Machine:
        """
        Converts a .jff (basic a XML) file into a Turing Machine. Some serious shit may happen if you manually edit the .jff file, don't do it.
        """
        tree = ET.parse(source)
        root = tree.getroot()
        automaton: ET.Element = root.find("./automaton")
        states: List[ET.Element] = automaton.findall("./state")
        transitions: List[ET.Element] = automaton.findall("./transition")

        states: List[State] = list(map(JFlapConverter.__parse_state__, states))

        transition_lambda: Callable[
            [ET.Element], Transition
        ] = lambda model: JFlapConverter.__parse_transition__(model, states)
        transitions: List[Transition] = list(map(transition_lambda, transitions))

        tape_amount = len(root.find("./automaton/transition").find("./read"))
        tapes = [Tape() for i in range(tape_amount)]

        return Machine(states=states, transitions=transitions, tapes=tapes)

    @classmethod
    def __parse_state__(cls, model: ET.Element) -> State:
        id = model.attrib["id"]
        label = model.attrib["name"]
        initial = len(model.findall("./initial")) > 0
        accept = len(model.findall("./final")) > 0
        return State(label, id, initial=initial, accept=accept)

    @classmethod
    def __parse_transition__(cls, model: ET.Element, states: List[State]) -> Transition:
        origin_id = model.find("./from").text
        destiny_id = model.find("./to").text

        origin = next(filter(lambda state: state.name == origin_id, states))
        destiny = next(filter(lambda state: state.name == destiny_id, states))
        reads = model.findall("./read")
        writes = model.findall("./write")
        moves = model.findall("./move")

        if len(reads) > 1:
            reads.sort(key=lambda item: item.attrib.get("tape"))
            writes.sort(key=lambda item: item.attrib.get("tape"))
            moves.sort(key=lambda item: item.attrib.get("tape"))

        reads = [item.text if item.text is not None else "_" for item in reads]
        writes = [item.text if item.text is not None else "_" for item in writes]
        moves = [Direction[item.text] for item in moves]

        return Transition(origin, destiny, reads, writes, moves)



import json

from autome.automatas.finite_automata import (
    State,
    Transition,
    DeterministicFiniteAutomata,
    NonDeterministicFiniteAutomata,
    transition,
)
from pathlib import Path
from xml.etree import ElementTree as ET
from typing import Callable, Dict, List, Union


class JFlapConverter:
    @classmethod
    def parse(
        cls, source: Path, deterministic=True
    ) -> Union[NonDeterministicFiniteAutomata, DeterministicFiniteAutomata]:
        """
        Converts a .jff file into a Finite State Automata. Some serious shit may happen if you manually edit the .jff file, don't do it.
        """
        tree = ET.parse(source)
        root = tree.getroot()
        automaton: ET.Element = root.find("./automaton")

        state_models: List[ET.Element] = automaton.findall("./state")
        states: List[State] = list(map(JFlapConverter.__parse_state__, state_models))

        transition_models: List[ET.Element] = automaton.findall("./transition")

        transitions = []
        for model in transition_models:
            transitions.append(JFlapConverter.__parse_transition__(model, states))

        if deterministic:
            return DeterministicFiniteAutomata(states=states, transitions=transitions)
        else:
            return NonDeterministicFiniteAutomata(
                states=states, transitions=transitions
            )

    @classmethod
    def save(cls, source: DeterministicFiniteAutomata, output_path: Path) -> bool:
        root = ET.Element("structure")
        type = ET.SubElement(root, "type")
        type.text = "fa"
        automaton = ET.SubElement(root, "automaton")

        for index, state in enumerate(source.states):
            _state = ET.SubElement(automaton, "state")
            _state.set("name", str(state.name))
            _state.set("id", str(state.id))
            ET.SubElement(_state, "x").text = f"{index * 30}"
            ET.SubElement(_state, "y").text = "50"

            if state.accept:
                ET.SubElement(_state, "final")

            if state.initial:
                ET.SubElement(_state, "initial")

        for index, transition in enumerate(source.transitions):
            _transition = ET.SubElement(automaton, "transition")
            ET.SubElement(_transition, "from").text = str(transition.origin.id)
            ET.SubElement(_transition, "to").text = str(transition.destiny.id)
            ET.SubElement(_transition, "read").text = transition.symbol

        document = ET.ElementTree(root)

        with open(output_path, "wb") as file:
            document.write(file, encoding="utf8", xml_declaration=True)
            return True

    @classmethod
    def __parse_state__(cls, model: ET.Element) -> State:
        label = model.attrib["id"]
        initial = len(model.findall("./initial")) > 0
        accept = len(model.findall("./final")) > 0
        return State(label, initial=initial, accept=accept)

    @classmethod
    def __parse_transition__(cls, model: ET.Element, states: List[State]) -> Transition:
        origin_id = model.find("./from").text
        destiny_id = model.find("./to").text

        origin = next(filter(lambda state: state.name == origin_id, states))
        destiny = next(filter(lambda state: state.name == destiny_id, states))
        reads = model.findall("./read")[0].text

        return Transition(origin, destiny, reads)


class JSONConverter:
    @classmethod
    def parse(
        cls, source: Union[Path, Dict], deterministic=True
    ) -> Union[NonDeterministicFiniteAutomata, DeterministicFiniteAutomata]:
        """
        Creates an instance of Machine parsed from a json file located at @path that follows the schematics provided by the project (see /machines folder)

        Throws FileNotFoundError if there's no file at @path
        Throws ValueError if the json file doesn't match the schematics
        """

        if isinstance(source, Dict):
            model = source
        elif isinstance(source, Path):
            with open(source, "r", encoding="utf8") as file:
                model = json.loads(file.read())

        states = [State.from_json(state) for state in model["states"]]
        transitions = [
            Transition.from_json(transition, states)
            for transition in model["transitions"]
        ]

        if deterministic:
            return DeterministicFiniteAutomata(states=states, transitions=transitions)
        else:
            return NonDeterministicFiniteAutomata(
                states=states, transitions=transitions
            )

    @classmethod
    def serialize(
        cls, source: Union[DeterministicFiniteAutomata, NonDeterministicFiniteAutomata]
    ) -> Dict:
        model = {}
        model["states"] = [state.to_json() for state in source.states]
        model["transitions"] = [
            transition.to_json() for transition in source.transitions
        ]

        return model

    @classmethod
    def save(
        cls,
        source: Union[DeterministicFiniteAutomata, NonDeterministicFiniteAutomata],
        output: Path,
    ):
        model = cls.serialize(source)

        with open(output, "w+") as fp:
            fp.write(json.dumps(model))

        print(f"Automato salvo em {output}")



from autome.automatas.finite_automata.state import State
from typing import Callable, Dict, List


class Transition:
    """
    A transition between two states in a Finite Automata. Any Transition has an origin state, a
    destiny origin and a trigger symbol.
    """

    def __init__(
        self,
        origin: State,
        destiny: State,
        symbol: str,
    ) -> None:
        self.origin = origin
        self.destiny = destiny
        self.symbol = symbol

    def __repr__(self):
        return f"Transition({self.origin.name} → {self.destiny.name}) : {self.symbol}"

    @classmethod
    def parse(cls, model: dict, states: List[State]) -> "Transition":
        """
        Returns a new instance of Transition based on the contents of @model. This function was written to be used within Machine.parse
        """
        find_origin: Callable[[State], str] = lambda state: state.uid == model["origin"]
        origin = next(filter(find_origin, states))

        find_destiny: Callable[[State], str] = (
            lambda state: state.uid == model["destiny"]
        )
        destiny = next(filter(find_destiny, states))

        return Transition(origin, destiny, model["symbol"])

    def to_json(self) -> Dict:
        return {
            "origin": self.origin.uid,
            "destiny": self.destiny.uid,
            "symbol": self.symbol,
        }

    @classmethod
    def from_json(cls, model: Dict, states: List[State]) -> "Transition":
        find_origin: Callable[[State], str] = lambda state: state.uid == model["origin"]
        origin = next(filter(find_origin, states))

        find_destiny: Callable[[State], str] = (
            lambda state: state.uid == model["destiny"]
        )
        destiny = next(filter(find_destiny, states))

        return Transition(origin, destiny, model["symbol"])

    def __eq__(self, other):
        return (
            self.origin == other.origin
            and self.destiny == other.destiny
            and self.symbol == other.symbol
        )

    def __hash__(self) -> int:
        concat = f"{self.destiny.__hash__()}.{self.origin.__hash__()}.{self.symbol.__hash__()}"
        return hash(concat)



from dataclasses import dataclass

from autome.regex.blocks import (
    SymbolAutomata,
    ConcatenationAutomata,
    UnionAutomata,
    KleeneAutomata,
    EpsilonAutomata,
)

from autome.automatas import NDFA


@dataclass
class ParserNode:
    def apply(self) -> "NDFA":
        raise NotImplementedError()


@dataclass
class SymbolNode(ParserNode):
    value: str

    def apply(self):
        # print(f'Entrou no symbol {self.value}')
        return SymbolAutomata(self.value) if self.value != "&" else EpsilonAutomata()

    def __repr__(self) -> str:
        return f"{self.value}"


@dataclass
class ConcatNode(ParserNode):
    node_a: any
    node_b: any

    def apply(self):
        # print(f'Entrou no concat {self.node_a} . {self.node_b}')
        return ConcatenationAutomata(self.node_a.apply(), self.node_b.apply())

    def __repr__(self) -> str:
        return f"({self.node_a} . {self.node_b})"


@dataclass
class UnionNode(ParserNode):
    node_a: any
    node_b: any

    def apply(self):
        # print(f'Entrou no union {self.node_a} | {self.node_b}')
        return UnionAutomata(self.node_a.apply(), self.node_b.apply())

    def __repr__(self) -> str:
        return f"({self.node_a} | {self.node_b})"


@dataclass
class KleeneClosureNode(ParserNode):
    node: any

    def apply(self):
        # print(f'Entrou no kleene {self.node}')
        return KleeneAutomata(self.node.apply())

    def __repr__(self) -> str:
        return f"({self.node})*"


@dataclass
class PositiveClosureNode(ParserNode):
    node: any

    def apply(self):
        return PositiveClosureNode(self.node.apply())

    def __repr__(self) -> str:
        return f"({self.node})+"



from os import stat
from typing import Dict, List
from itertools import count
from time import time
from uuid import uuid4


class State:
    _ids = count(0)

    """
    A possible State of a Finite Automata.
    """

    def __init__(
        self,
        name: str = None,
        initial=False,
        accept=False,
        type=None,
        uid=None,
        parts=[],
    ) -> None:
        self.initial: bool = initial
        self.accept: bool = accept
        self.id = next(self._ids)
        self.type: str = type
        self.parts: List[float] = parts

        if uid is not None:
            self.uid = uid
        else:
            self.uid = str(uuid4())

        if name is None:
            self.name = f"q{self.id}"
        else:
            self.name: str = name

    @classmethod
    def parse(cls, model: dict) -> "State":
        """
        Returns a new instance of State based on the contents of @model. This function was written to be used within Machine.parse
        """
        return State(
            model.get("name"),
            initial=model.get("initial"),
            accept=model.get("accept"),
            uid=model.get("uid"),
        )

    def to_json(self) -> Dict:
        return {
            "name": self.name,
            "initial": self.initial,
            "accept": self.accept,
            "uid": self.uid,
            "type": self.type,
        }

    @classmethod
    def from_json(cls, model: Dict) -> "State":
        return State(
            initial=model.get("initial"),
            accept=model.get("accept"),
            uid=model.get("uid"),
            type=model.get("type"),
        )

    @classmethod
    def join(cls, states: List["State"]) -> "State":
        if len(states) == 1:
            return states[0]

        parts = list(map(lambda state: str(state.uid), states))
        parts.sort()

        names = list(map(lambda state: state.name, states))
        names.sort()
        name = ".".join(names)

        accept = len(list(filter(lambda state: state.accept, states))) > 0
        initial = len(list(filter(lambda state: state.initial, states))) > 0
        types = set(map(lambda state: state.type, states))
        types = list(filter(lambda t: t is not None, types))

        if len(types) > 0:
            type = types[0]
        else:
            type = None

        return State(name=name, accept=accept, initial=initial, type=type, parts=parts)

    def debug(self):
        return f"{self.name} ({self.type})"

    def __repr__(self):
        return f"q{self.id}"
        # return f"State(id: {self.name}, accept: {self.accept}, initial: {self.initial}, type: {self.type})"

    def __lt__(self, other: "State") -> bool:
        return self.name < other.name

    def __eq__(self, other):
        if type(other) != State:
            return False

        # Problema: criar uma forma de identificação para estados que atenda os requisitos:
        #   - O identificador deve ser globalmente único, para permitir save/parse
        #   - Ao criar estados com State.join(), sempre que fizer join(a,b,c) o estado resultante deve possuir o mesmo identificador

        # if (len(self.parts) > 0 and len(other.parts) > 0):
        # print(self.parts)
        # print(other.parts)
        # print(self.parts == other.parts)
        # return self.parts == other.parts

        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)



from dataclasses import dataclass
from typing import Optional
from autome.regex.regex import Regex


@dataclass
class Definition:
    name: str
    expression: str
    regex: Optional[Regex] = None

    def __repr__(self):
        return f"{self.name} => {self.expression}"

    def dict(self):
        return {"name": self.name, "expression": self.expression}


@dataclass
class Token:
    type: str
    value: str

    def __repr__(self) -> str:
        return f"{self.type}:{self.value}"



from autome.utils.enums import Direction
from autome.automatas.turing_machine.state import State
from typing import Callable, List


class Transition:
    """
    A transition between two states in a Turing-Machine. Any Transition has an origin state, a
    """

    def __init__(
        self,
        origin: State,
        destiny: State,
        reads: List[str],
        writes: List[str],
        moves: List[Direction],
    ) -> None:
        self.origin = origin
        self.destiny = destiny
        self.reads = reads
        self.writes = writes
        self.moves = moves

    def __repr__(self):
        return f"Transition({self.origin.label} → {self.destiny.label}) : {self.reads} → {self.writes}"

    @classmethod
    def parse(cls, model: dict, states: List[State]) -> "Transition":
        """
        Returns a new instance of Transition based on the contents of @model. This function was written to be used within Machine.parse
        """
        find_origin: Callable[[State], str] = (
            lambda state: state.name == model["origin"]
        )
        origin = next(filter(find_origin, states))

        find_destiny: Callable[[State], str] = (
            lambda state: state.name == model["destiny"]
        )
        destiny = next(filter(find_destiny, states))

        moves = [Direction[dir.upper()] for dir in model["moves"]]

        return Transition(origin, destiny, model["reads"], model["writes"], moves)



from autome.automatas.turing_machine.state import State
from autome.automatas.turing_machine.tape import Tape
from autome.automatas.turing_machine.transition import Transition
from typing import List, Callable, Deque
from pathlib import Path
from collections import deque
import json


class BaseMachine:
    """
    Generic machine, accepts states, tapes and transitions as inputs.

    Outputs boolean for result upon transition to acceptance state or reject state.

    A machine can only have one initial state, but it may have multiple acceptance and reject states. Computation will be stopped as soon as the machine reaches one of those two types of state.
    """

    def __init__(
        self, states=[], tapes=[], transitions=[], branches: Deque = None
    ) -> None:
        self.states: List[State] = states
        self.current_state = next(filter(lambda state: state.initial, self.states))
        self.transitions: List[Transition] = transitions
        self.output = False

        if branches is not None:
            self.branches = branches
        else:
            self.branches: Deque[BaseMachine] = deque([self])

    def __repr__(self):
        return f"BaseMachine(states: {len(self.states)})"

    @classmethod
    def parse(cls, path: Path) -> "BaseMachine":
        """
        Creates an instance of Machine parsed from a json file located at @path that follows the schematics provided by the project (see /machines folder)

        Throws FileNotFoundError if there's no file at @path
        Throws ValueError if the json file doesn't match the schematics
        """
        with open(path, "r", encoding="utf8") as file:
            model = json.loads(file.read())
            tapes = [Tape.parse(tape) for tape in model["tapes"]]
            states = [State.parse(state) for state in model["states"]]
            transitions = [
                Transition.parse(transition, states)
                for transition in model["transitions"]
            ]

        return BaseMachine(states=states, tapes=tapes, transitions=transitions)

    def clone(self) -> "BaseMachine":
        """
        Returns a clone of the calling Turing Machine. The states, transitions and branches will be shallow copies
        while the tapes are clones with differente memory spaces.

        States and Transitions are read-only, so a shallow copy is fine.

        Tapes have write and move operations, so they must be new objects on the new machine.

        The branches must be shared between instances (we call it a "gambiarra")
        """
        tapes = [tape.clone() for tape in self.tapes]

        return BaseMachine(
            states=self.states,
            tapes=tapes,
            transitions=self.transitions,
            branches=self.branches,
        )



from autome.automatas.finite_automata.machine import NonDeterministicFiniteAutomata
from autome.automatas.finite_automata.state import State
from autome.automatas.finite_automata.transition import Transition


class ConcatenationAutomata(NonDeterministicFiniteAutomata):
    """Creates a concatenation recognizer automata, which accepts the concatenations of the languages regognized by its inputs.

    Args:
        a (NonDeterministicFiniteAutomata): first operand.
        b (NonDeterministicFiniteAutomata): second operand.
    """

    def __init__(
        self, a: NonDeterministicFiniteAutomata, b: NonDeterministicFiniteAutomata
    ) -> "ConcatenationAutomata":
        # Using deep copys to avoid modifying the original automata.
        temp_a = a.clone()
        temp_b = b.clone()

        new = []
        # All transitions going from the initial state of the second DFA now should start on the
        # final state of the first DFA
        for transition in temp_b.transitions:
            if transition.origin != temp_b.initial():
                continue

            for state in temp_a.final():
                # Handles the case where a initial state has transitions to itself
                destiny = (
                    transition.destiny
                    if transition.destiny != temp_b.initial()
                    else state
                )
                new.append(Transition(state, transition.destiny, transition.symbol))

        for transition in temp_b.transitions:
            if transition.origin != temp_b.initial():
                continue
            temp_b.transitions.remove(transition)

        # Now we can remove the initial state of the second DFA
        temp_b.states.remove(temp_b.initial())

        # And the final states from the first DFA are no longer final
        for state in temp_a.final():
            # If a initial state is also final, probably is for accepting the empty string
            state.accept = False

        states = temp_a.states + temp_b.states
        transitions = temp_a.transitions + temp_b.transitions + new

        super().__init__(states, transitions)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autome',
 'autome.automatas',
 'autome.automatas.finite_automata',
 'autome.automatas.turing_machine',
 'autome.grammars',
 'autome.regex',
 'autome.regex.blocks',
 'autome.utils']

package_data = \
{'': ['*']}

install_requires = \
['black>=22.3.0,<23.0.0',
 'click>=8.1.3,<9.0.0',
 'ksuid>=1.3,<2.0',
 'tabulate>=0.8.9,<0.9.0']

setup_kwargs = {
    'name': 'autome',
    'version': '0.2.0',
    'description': 'Automata simulation with Python',
    'long_description': None,
    'author': 'João Vitor Maia',
    'author_email': 'maia.tostring@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from autome.regex.blocks.concat import ConcatenationAutomata
from autome.regex.blocks.kleene import KleeneAutomata
from autome.regex.blocks.symbol import SymbolAutomata
from autome.regex.blocks.epsilon import EpsilonAutomata
from autome.regex.blocks.union import UnionAutomata



from autome.automatas import NDFA
from autome.regex import Lexer, Parser, Interpreter


class Regex:
    def __init__(self, expression) -> None:
        self.expression = expression
        self.lexer = Lexer(self.expression)
        self.tokens = self.lexer.generate_tokens()
        self.parser = Parser(self.tokens)
        self.tree = self.parser.parse()
        self.interpreter = Interpreter()

    def match(self, test) -> bool:
        return self.interpreter.run(self.tree).determinize().accepts(test)

    def automata(self) -> NDFA:
        return self.interpreter.run(self.tree)



from typing import List


class State:
    """
    A possible State of a Turing-Machine.
    """

    def __init__(
        self, label: str, name: str, initial=False, accept=False, reject=False
    ) -> None:
        self.label = label
        self.name = name
        self.initial = initial
        self.accept = accept
        self.reject = reject

    @classmethod
    def parse(cls, model: dict) -> "State":
        """
        Returns a new instance of State based on the contents of @model. This function was written to be used within Machine.parse
        """
        return State(
            model["label"],
            model["name"],
            initial=model["initial"],
            accept=model["accept"],
            reject=model["reject"],
        )

    def __repr__(self):
        return f"State(label: {self.label}, accept: {self.accept}, reject: {self.reject}, initial: {self.initial})"



from enum import Enum
from itertools import chain, combinations


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return set(chain.from_iterable(combinations(s, r) for r in range(len(s) + 1)))
